import os
import re
import base64
import mimetypes
from io import BytesIO
from .base import LottieObject, LottieProp, PseudoBool, Index
from .layers import Layer
from .shapes import ShapeElement
from .composition import Composition


## @ingroup Lottie
class Asset(LottieObject):
    @classmethod
    def _load_get_class(cls, lottiedict):
        if "p" in lottiedict or "u" in lottiedict:
            return Image
        if "layers" in lottiedict:
            return Precomp


## @ingroup Lottie
class Image(Asset):
    """!
        External image

        @see http://docs.aenhancers.com/sources/filesource/
    """
    _props = [
        LottieProp("height", "h", float, False),
        LottieProp("width", "w", float, False),
        LottieProp("id", "id", str, False),
        LottieProp("image", "p", str, False),
        LottieProp("image_path", "u", str, False),
        LottieProp("is_embedded", "e", PseudoBool, False),
    ]

    @staticmethod
    def guess_mime(file):
        if isinstance(file, str):
            filename = file
        elif hasattr(file, "name"):
            filename = file.name
        else:
            return "application/octet-stream"
        return mimetypes.guess_type(filename)

    def __init__(self, id=""):
        ## Image Height
        self.height = 0
        ## Image Width
        self.width = 0
        ## Image ID
        self.id = id
        ## Image name
        self.image = ""
        ## Image path
        self.image_path = ""
        ## Image data is stored as a data: url
        self.is_embedded = False

    def load(self, file, format=None):
        """!
        @param file     Filename, file object, or PIL.Image.Image to load
        @param format   Format to store the image data as
        """
        from PIL import Image

        if not isinstance(file, Image.Image):
            image = Image.open(file)
        else:
            image = file

        self._id_from_file(file)

        self.image_path = ""
        if format is None:
            format = (image.format or "png").lower()
        self.width, self.height = image.size
        output = BytesIO()
        image.save(output, format=format)
        self.image = "data:image/%s;base64,%s" % (
            format,
            base64.b64encode(output.getvalue()).decode("ascii")
        )
        self.is_embedded = True
        return self

    def _id_from_file(self, file):
        if not self.id:
            if isinstance(file, str):
                self.id = os.path.basename(file)
            elif hasattr(file, "name"):
                self.id = os.path.basename(file.name)
            elif hasattr(file, "filename"):
                self.id = os.path.basename(file.filename)
            else:
                self.id = "image_%s" % id(self)

    @classmethod
    def embedded(cls, image, format=None):
        """!
        Create an object from an image file
        """
        lottie_image = cls()
        return lottie_image.load(image, format)

    @classmethod
    def linked(cls, filename):
        from PIL import Image
        image = Image.open(filename)
        lottie_image = cls()
        lottie_image._id_from_file(filename)
        lottie_image.image_path, lottie_image.image = os.path.split(filename)
        lottie_image.image_path += "/"
        lottie_image.width = image.width
        lottie_image.height = image.height
        return lottie_image

    def image_data(self):
        """
        Returns a tuple (format, data) with the contents of the image

        `format` is a string like "png", and `data` is just raw binary data.

        If it's impossible to fetch this info, returns (None, None)
        """
        if self.is_embedded:
            m = re.match("data:[^/]+/([^;,]+);base64,(.*)", self.image)
            if m:
                return m.group(1), base64.b64decode(m.group(2))
            return None, None
        path = self.image_path + self.image
        if os.path.isfile(path):
            with open(path, "rb") as imgfile:
                return os.path.splitext(path)[1][1:], imgfile.read()
        return None, None


## @ingroup Lottie
class CharacterData(LottieObject):
    """!
    Character shapes
    """
    _props = [
        LottieProp("shapes", "shapes", ShapeElement, True),
    ]

    def __init__(self):
        self.shapes = []


## @ingroup Lottie
class Chars(LottieObject):
    """!
    Defines character shapes to avoid loading system fonts
    """
    _props = [
        LottieProp("character", "ch", str, False),
        LottieProp("font_family", "fFamily", str, False),
        LottieProp("font_size", "size", float, False),
        LottieProp("font_style", "style", str, False),
        LottieProp("width", "w", float, False),
        LottieProp("data", "data", CharacterData, False),
    ]

    def __init__(self):
        ## Character Value
        self.character = ""
        ## Character Font Family
        self.font_family = ""
        ## Character Font Size
        self.font_size = 0
        ## Character Font Style
        self.font_style = "" # Regular
        ## Character Width
        self.width = 0
        ## Character Data
        self.data = CharacterData()

    @property
    def shapes(self):
        return self.data.shapes


## @ingroup Lottie
class Precomp(Asset, Composition):
    _props = [
        LottieProp("id", "id", str, False),
    ]

    def __init__(self, id="", animation=None):
        super().__init__()
        ## Precomp ID
        self.id = id
        self.animation = animation
        if animation:
            self.animation.assets.append(self)

    def _on_prepare_layer(self, layer):
        if self.animation:
            self.animation.prepare_layer(layer)

    def set_timing(self, outpoint, inpoint=0, override=True):
        for layer in self.layers:
            if override or layer.in_point is None:
                layer.in_point = inpoint
            if override or layer.out_point is None:
                layer.out_point = outpoint



# SPDX-FileCopyrightText: 2019-2023 Mattia Basaglia <dev@dragon.best>
# SPDX-License-Identifier: GPL-3.0-or-later
import os
from PIL import Image
from PIL import features


def _png_gif_prepare(image):
    if image.mode not in ["RGBA", "RGBa"]:
        image = image.convert("RGBA")
    alpha = image.getchannel("A")
    image = image.convert("RGB").convert('P', palette=Image.ADAPTIVE, colors=255)
    mask = Image.eval(alpha, lambda a: 255 if a <= 128 else 0)
    image.paste(255, mask=mask)
    return image


def save_gif(window, document, comp, file, name, import_export, settings):
    skip_frames = 1
    start = int(comp.animation.first_frame)
    end = int(comp.animation.last_frame)

    import_export.progress_max_changed(end-start)

    frames = []
    for i in range(start, end+1, skip_frames):
        import_export.progress(i-start)
        frames.append(_png_gif_prepare(comp.render_image(i)))

    duration = int(round(1000 / comp.fps * skip_frames / 10)) * 10
    frames[0].save(
        file,
        format='GIF',
        append_images=frames[1:],
        save_all=True,
        duration=duration,
        loop=0,
        transparency=255,
        disposal=2,
    )


def save_webp(window, document, comp, file, name, import_export, settings):
    if not features.check("webp_anim"):
        window.warning("WebP animations not supported in this system")
        return

    skip_frames = settings["skip_frames"]
    dpi = 96

    start = int(comp.animation.first_frame)
    end = int(comp.animation.last_frame)
    import_export.progress_max_changed(end-start)

    frames = []
    for i in range(start, end+1, skip_frames):
        import_export.progress(i-start)
        frames.append(comp.render_image(i))

    duration = int(round(1000 / comp.fps * skip_frames))
    frames[0].save(
        file,
        format='WebP',
        append_images=frames[1:],
        save_all=True,
        duration=duration,
        loop=0,
        background=(0, 0, 0, 0),
        lossless=settings["lossless"],
        quality=settings["quality"],
        method=settings["method"]
    )


def open_image(window, document, file, name, import_export, settings):
    raster = Image.open(file)
    comp = document.assets.add_composition()
    comp.width = raster.width
    comp.height = raster.height

    if not hasattr(raster, "is_animated"):
        raster.n_frames = 1
        raster.seek = lambda x: None

    parent_layer = comp.add_shape("Layer")
    parent_layer.name = comp.name = os.path.basename(name)
    time = 0

    for frame in range(raster.n_frames):
        layer = parent_layer.add_shape("Layer")
        layer.name = "Frame %s Layer" % frame
        layer.animation.first_frame = time

        raster.seek(frame)

        if "duration" in raster.info:
            time += raster.info.get("duration") / 1000 * comp.fps
        else:
            time += 1
        layer.animation.last_frame = time

        asset = document.assets.add_image(raster)

        shape = layer.add_shape("Image")
        shape.image = asset
        shape.name = "Frame %s" % frame

    comp.animation.last_frame = parent_layer.animation.last_frame = time



# SPDX-FileCopyrightText: 2019-2023 Mattia Basaglia <dev@dragon.best>
# SPDX-License-Identifier: GPL-3.0-or-later
def main(window, document, settings):
    window.warning(
        settings["greeting"]
    )



# SPDX-FileCopyrightText: 2019-2023 Mattia Basaglia <dev@dragon.best>
# SPDX-License-Identifier: GPL-3.0-or-later
import glaxnimate


class Replacer(glaxnimate.model.Visitor):
    def __init__(self, color_search, color_replace):
        super().__init__()
        self.color_search = color_search
        self.color_replace = color_replace

    def on_visit_node(self, node):
        if (
            isinstance(node, glaxnimate.model.shapes.Styler) and
            node.color.value == self.color_search
        ):
            node.color.value = self.color_replace


def main(window, document, settings):
    dialog = window.create_dialog("replace.ui")
    if not dialog:
        return

    dialog.set_value("search", "color", window.fill_color)
    if dialog.exec():
        with document.macro("Replace color"):
            Replacer(dialog.get_value("search", "color"), dialog.get_value("replace", "color")).visit(document, True)



# SPDX-FileCopyrightText: 2019-2023 Mattia Basaglia <dev@dragon.best>
# SPDX-License-Identifier: GPL-3.0-or-later
import io
import json
import gzip

from lottie.parsers.sif import parse_sif_file
from lottie.parsers.sif.builder import to_sif
from lottie.objects.animation import Animation

import glaxnimate


def save_synfig(window, document, file, fname, import_export, settings):
    lottie_str = glaxnimate.io.registry["lottie"].save(document)
    lottie_dict = json.loads(lottie_str)
    animation = Animation.load(lottie_dict)
    dom = to_sif(animation).to_xml()
    compressed = fname.endswith("sifz")
    fp = io.StringIO()
    dom.writexml(fp, "", "  ", "\n", "utf-8")
    if compressed:
        with gzip.open(file, "wb") as gzfile:
            gzfile.write(fp.getvalue().encode("utf-8"))
    else:
        file.write(fp.getvalue().encode("utf-8"))


def open_synfig(window, document, file, fname, import_export, settings):
    animation = parse_sif_file(file)
    lottie_str = json.dumps(animation.to_dict()).encode("utf-8")
    glaxnimate.io.registry["lottie"].load(document, lottie_str)



# SPDX-FileCopyrightText: 2019-2023 Mattia Basaglia <dev@dragon.best>
# SPDX-License-Identifier: GPL-3.0-or-later
import io
import re
import os
import json
import string
import base64
import zipfile
from PIL import Image

import glaxnimate


def asset_data(asset):
    if asset["e"]:
        m = re.match("data:[^/]+/([^;,]+);base64,(.*)", asset["p"])
        if not m:
            return None, None
        return m.group(1), base64.b64decode(m.group(2))

    path = os.path.join(asset["u"], asset["p"])
    if not os.path.isfile(path):
        return None, None

    with open(path, "rb") as imgfile:
        return os.path.splitext(path)[1][1:], imgfile.read()


def save_dotlottie(window, document, file, fname, import_export, settings):
    files = {}

    if settings["append"]:
        with zipfile.ZipFile(fname, "r") as zf:
            with zf.open("manifest.json") as manifest:
                meta = json.load(manifest)
                meta["custom"].update(document.metadata)

            for name in zf.namelist():
                if name != "manifest.json":
                    files[name] = zf.read(name)
    else:
        meta = {
            "generator": "Glaxnimate " + glaxnimate.__version__,
            "version": 1.0,
            "revision": 1,
            "author": str(document.metadata.get("author", "")),
            "animations": [],
            "custom": document.metadata,
            "author": document.info.author,
        }

    try:
        meta["revision"] = int(document.metadata["revision"])
    except (KeyError, ValueError):
        pass

    id = settings["id"]
    if not id:
        if document.main.name:
            idok = string.ascii_letters + string.digits + "_-"
            id = "".join(filter(lambda x: x in idok, document.main.name.replace(" ", "_")))
        if not id:
            id = "animation_%s" % len(meta["animations"])

    meta["animations"].append({
        "id": id,
        "speed": settings["speed"],
        "themeColor": settings["theme_color"],
        "loop": settings["loop"],
    })

    lottie_str = glaxnimate.io.registry["lottie"].save(document)
    lottie_dict = json.loads(lottie_str)

    if settings["pack_assets"]:
        image_no = 0
        for asset in lottie_dict.get("assets", []):
            if "e" not in asset or "p" not in asset:
                continue

            ext, data = asset_data(asset)
            if not ext:
                continue

            pathname = "images/"
            while True:
                basename = "image_%s.%s" % (image_no, ext)
                image_no += 1
                if pathname+basename not in files:
                    break
            files[pathname+basename] = data
            asset["u"] = pathname
            asset["p"] = basename
            asset["e"] = 0

    files["manifest.json"] = json.dumps(meta)
    files["animations/%s.json" % id] = json.dumps(lottie_dict)

    with open(fname, "wb") as fd:
        with zipfile.ZipFile(fd, "w") as zf:
            for name, data in files.items():
                zf.writestr(name, data)


def load_asset(asset, file):
    image = Image.open(file)
    asset["u"] = ""
    asset["p"] = ""
    format = (image.format or "png").lower()
    asset["w"], asset["h"] = image.size
    output = io.BytesIO()
    image.save(output, format=format)
    asset["p"] = "data:image/%s;base64,%s" % (
        format,
        base64.b64encode(output.getvalue()).decode("ascii")
    )
    asset["e"] = 1

def open_dotlottie(window, document, file, fname, import_export, settings):
    file.open("rb")
    with zipfile.ZipFile(file, "r") as zf:
        with zf.open("manifest.json") as manifest:
            meta = json.load(manifest)

        animations = {}
        ids = {}
        for anim in meta["animations"]:
            anim_id = anim["id"]
            info = zf.getinfo("animations/%s.json" % anim_id)
            with zf.open(info) as animfile:
                lottie = json.load(animfile)

            animations[anim_id] = lottie
            ids[lottie.get("nm", anim_id)] = anim_id

        chosen_id = None
        if len(meta["animations"]) > 1:
            chosen_id = window.choose_option("Animation", ids, "Open dotLottie")
        if chosen_id is None:
            chosen_id = meta["animations"][0]["id"]

        lottie = animations[chosen_id]

        for asset in lottie.get("assets", []):
            # .lottie is very werid on how it handles assets so we can't
            # rely on `e` or `p`...
            if asset.get("p", "").startswith("data:"):
                continue

            asset_fname = os.path.join(asset["u"], asset["p"])
            if asset_fname not in zf.namelist() and "images/" + asset_fname in zf.namelist():
                asset_fname = "images/" + asset_fname

            if asset_fname in zf.namelist():
                with zf.open(asset_fname) as asset_file:
                    load_asset(asset, asset_file)

        glaxnimate.io.registry["lottie"].load(document, json.dumps(lottie).encode("utf-8"))

        custom_meta = meta.get("custom", {})
        if "author" in meta:
            custom_meta["author"] = meta["author"]
        if "revision" in meta:
            custom_meta["revision"] = meta["revision"]
        document.metadata = custom_meta



from .base import LottieObject, LottieProp, PseudoBool, Index
from .layers import Layer
from .assets import Asset, Chars, Precomp
from .text import FontList
from .composition import Composition

##\defgroup Lottie Lottie
#
# Objects of the lottie file structure.

## \defgroup LottieCheck Lottie (to check)
#
# Lottie objects that have not been tested


## @ingroup Lottie
class Animation(Composition):
    """!
    Top level object, describing the animation

    @see http://docs.aenhancers.com/items/compitem/
    """
    _props = [
        LottieProp("version", "v", str, False),
        LottieProp("frame_rate", "fr", float, False),
        LottieProp("in_point", "ip", float, False),
        LottieProp("out_point", "op", float, False),
        LottieProp("width", "w", int, False),
        LottieProp("height", "h", int, False),
        LottieProp("name", "nm", str, False),
        LottieProp("threedimensional", "ddd", PseudoBool, False),
        LottieProp("assets", "assets", Asset, True),
        #LottieProp("comps", "comps", Animation, True),
        LottieProp("fonts", "fonts", FontList),
        LottieProp("chars", "chars", Chars, True),
        #LottieProp("markers", "markers", Marker, True),
        #LottieProp("motion_blur", "mb", MotionBlur, False),
    ]
    _version = "5.5.2"

    def __init__(self, n_frames=60, framerate=60):
        super().__init__()
        ## The time when the composition work area begins, in frames.
        self.in_point = 0
        ## The time when the composition work area ends.
        ## Sets the final Frame of the animation
        self.out_point = n_frames
        ## Frames per second
        self.frame_rate = framerate
        ## Composition Width
        self.width = 512
        ## Composition has 3-D layers
        self.threedimensional = False
        ## Composition Height
        self.height = 512
        ## Bodymovin Version
        self.version = self._version
        ## Composition name
        self.name = None
        ## source items that can be used in multiple places. Comps and Images for now.
        self.assets = [] # Image, Precomp
        ## source chars for text layers
        self.chars = None
        ## Available fonts
        self.fonts = None

    def precomp(self, name):
        for ass in self.assets:
            if isinstance(ass, Precomp) and ass.id == name:
                return ass
        return None

    def _on_prepare_layer(self, layer):
        if layer.in_point is None:
            layer.in_point = self.in_point
        if layer.out_point is None:
            layer.out_point = self.out_point

    def tgs_sanitize(self):
        """!
        Cleans up some things to ensure it works as a telegram sticker
        """
        if self.width != 512 or self.height != 512:
            scale = min(512/self.width, 512/self.height)
            self.width = self.height = 512

            for layer in self.layers:
                if layer.parent_index:
                    continue

                if layer.transform.scale.animated:
                    for kf in layer.transform.scale.keyframes:
                        if kf.start is not None:
                            kf.start *= scale
                        if kf.end is not None:
                            kf.end *= scale
                else:
                    layer.transform.scale.value *= scale

                if layer.transform.position.animated:
                    for kf in layer.transform.position.keyframes:
                        if kf.start is not None:
                            kf.start *= scale
                        if kf.end is not None:
                            kf.end *= scale
                else:
                    layer.transform.position.value *= scale

        if self.frame_rate < 45:
            self.frame_rate = 30
        else:
            self.frame_rate = 60

    def _fixup(self):
        super()._fixup()
        if self.assets:
            for ass in self.assets:
                if isinstance(ass, Precomp):
                    ass.animation = self
                    ass._fixup()

    def __str__(self):
        return self.name or super().__str__()
